#ifndef ZAD2_QUAD_H
#define ZAD2_QUAD_H
#include <string>
#include <utility>

#include "../Info.h"
#include "Variable.h"
using namespace std;

static const std::string opNames[] = { "_U", "-", "!", "", "_UE",
                                       "_B", "+", "-", "*", "/", "%", "&", "|", "^", "_BE",
                                       "_C", "<", "<=", "==", "!=", ">=", ">", "_CE"};

class Quadruple : std::enable_shared_from_this<Quadruple> {
public:
	class Op {
	public:
		enum OpType {
			UNARY,
			Neg, Not, Copy, UNARY_END,
			BINARY,
			Plus, Minus, Mul, Div, Mod, And, Or, Xor, BINARY_END,
			CMP,
			LT, LE, EQ, NEQ, GE, GT, CMP_END
		};
		OpType op;

		Op(OpType op = UNARY) : op(op) {};
		operator OpType() const { return op; };

		static OpType kind(OpType op) {
			if (op <= UNARY_END) return UNARY;
			if (op <= BINARY_END) return BINARY;
			return CMP;
		};
		static const std::string& name(OpType op) {
			return opNames[op];
		}
	};

	virtual ~Quadruple() {};
	virtual std::string toString() const = 0;
};

class QWriteVar : public Quadruple {
public:
	VariablePtr loc;
	explicit QWriteVar(VariablePtr loc) : Quadruple(), loc(std::move(loc)) {};

	std::string toString() const override { return loc ? ("\t" + loc->name + " := ") : "\t"; }
};

class QLabel : public Quadruple {
public:
	string label;
	std::string toString() const override { return "\"" + label + "\":"; }
	QLabel(std::string comment) : label(comment) {}
};

class QAssign : public QWriteVar {
public:
	Op op;
	vector<VariablePtr> args;

	QAssign(VariablePtr loc, Op op, vector<VariablePtr> list) : QWriteVar(std::move(loc)), op(op), args(std::move(list)) {};
	QAssign(VariablePtr loc, VariablePtr l, Op op, VariablePtr r) : QWriteVar(std::move(loc)), op(op), args({std::move(l), std::move(r)}) {};
	QAssign(VariablePtr loc, Op op, VariablePtr x) : QWriteVar(std::move(loc)), op(op), args({std::move(x)}) {};

	std::string toString() const override {
		if (args.size() == 1)
			return QWriteVar::toString() + Op::name(op) + " " + args[0]->name;
		else
			return QWriteVar::toString() + args[0]->name + " " + Op::name(op) + " " + args[1]->name;
	}
};

class QJump : public Quadruple {
public:
	shared_ptr<QLabel> target;
	explicit QJump(shared_ptr<QLabel> target) : target(std::move(target)) {};

	std::string toString() const override {
		return "\tjump " + target->label;
	}
};

class QJumpCond : public QJump {
public:
	VariablePtr left;
	Op op;
	VariablePtr right;
	QJumpCond(shared_ptr<QLabel> to, VariablePtr left, Op op, VariablePtr right)
			: QJump(std::move(to)), left(std::move(left)), op(op), right(std::move(right)) {};

	QJumpCond(shared_ptr<QLabel> to, Op op, VariablePtr right)
			: QJumpCond(std::move(to), nullptr, op, std::move(right)) {};

	std::string toString() const override {
		if (!left)
			return "\tjump \"" + target->label + "\" if " + Op::name(op) + " " + right->name;
		else
			return "\tjump \"" + target->label + "\" if " + left->name + " " + Op::name(op) + " " + right->name;
	}
};

class QParam : public Quadruple {
public:
	VariablePtr param;
	int num;

	QParam(VariablePtr param, int num) : param(std::move(param)), num(num) {};

	std::string toString() const override {
		if (num == 0) {
			return "\tparam self " + param->name;
		} else {
			return "\tparam #" + to_string(num) + " <- " + param->name;
		}
	}
};

class QCall : public QWriteVar {
public:
	FunctionInfoPtr fn;
	int params;
	VariablePtr self;
	QCall(VariablePtr loc, FunctionInfoPtr fn, int params, VariablePtr self)
		: QWriteVar(std::move(loc)), fn(std::move(fn)), params(params), self(std::move(self)) {};

	std::string toString() const override {
		if (self) {
			return QWriteVar::toString() + "call " + fn->name + " of " + self->name + " with " + to_string(params) + " params";
		} else {
			return QWriteVar::toString() + "call " + fn->name + " with " + to_string(params) + " params";
		}
	}
};

class QReturn : public Quadruple {
public:
	VariablePtr ret;
	explicit QReturn(VariablePtr ret) : ret(std::move(ret)) {};

	std::string toString() const override {
		if (ret) {
			return "\treturn " + ret->name;
		} else {
			return "\treturn";
		}
	}
};

class QAccess : public QWriteVar {
public:
	QAccess(VariablePtr loc, VariablePtr base, VariablePtr index, int multiplier, int offset)
		: QWriteVar(std::move(loc)), base(std::move(base)), index(std::move(index)), multiplier(multiplier), offset(offset)
	{};
	// leal    8(,%eax,4), %eax         # Arithmetic: multiply eax by 4 and add 8
	// leal    (%edx,%eax,2), %eax      # Arithmetic: multiply eax by 2 and add edx
	VariablePtr base;
	VariablePtr index;
	int multiplier;
	int offset;

	std::string toString() const override {
		auto ret = QWriteVar::toString() + "[" + base->name;
		if (multiplier || !(index->constExpr || !index->val)) ret += " + " + to_string(multiplier) + "*" + index->name;
		if (offset) ret += " + " + to_string(offset);
		return ret + "]";
	}
};

class QAlloc : public QWriteVar {
public:
	size_t size;
	VariablePtr count;
	std::string virtSymbol;
	QAlloc(VariablePtr loc, size_t size, std::string virtSymbol, VariablePtr count = nullptr)
		: QWriteVar(std::move(loc)), size(size), count(std::move(count)), virtSymbol(std::move(virtSymbol)) {};

	std::string toString() const override {
		std::string ret = QWriteVar::toString() + "alloc (" + to_string(size) + ")";
		if (count) { ret += " [" + count->name + "]"; }
		if (!virtSymbol.empty()) { ret += " " + virtSymbol; }
		return ret;
	}
};


#endif //ZAD2_QUAD_H
